/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.client;

import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.pua.prot.client.dataobject.Column;
import de.bsvrz.pua.prot.client.dataobject.ProtocolAdministrationData;
import de.bsvrz.pua.prot.client.dataobject.ProtocolAggregationData;
import de.bsvrz.pua.prot.client.dataobject.ProtocolAggregationResultData;
import de.bsvrz.pua.prot.client.dataobject.ProtocolNoChanges;
import de.bsvrz.pua.prot.client.dataobject.ProtocolResult;
import de.bsvrz.pua.prot.client.dataobject.ProtocolResultData;
import de.bsvrz.pua.prot.util.ProcessingInformation.ApplyAggregation;
import de.bsvrz.sys.funclib.losb.util.Util;

import javax.swing.*;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableColumn;
import java.awt.*;
import java.util.*;
import java.util.List;

/**
 * Zeigt ein Protokolldatensätze in einem Fenster an.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ProtocolViewer extends JPanel {

	private static final long serialVersionUID = -8347368586876630163L;

	private final DataModel _dataModel;

	/** Verwaltet Datensätze */
	ProtocolTableModel _model = null;

	/** Fenster */
	JFrame _frame = null;

	/** Protokoll-Id */
	private Long _protocolId = null;

	/** Skriptname */
	private String _scriptName = null;

	/** Zeitpunkt der Protokollanfroderung (grob) */
	private String _date = Util.msToDate(System.currentTimeMillis());

	/** Konstruktor.
	 * @param dataModel DataModel*/
	public ProtocolViewer(final DataModel dataModel) {
		super(new GridLayout(1, 0));
		_dataModel = dataModel;
	}

	/**
	 * Protokolldatensatz hinzufügen. Beim ersten Aufruf muss ein Objekt der Klasse {@link ProtocolAdministrationData} übergeben werden, um die Tabelle zu
	 * initialisieren.
	 *
	 * @param pr Protokolldatensatz
	 */
	synchronized public void add(ProtocolResult pr) {
		if(_frame == null && pr instanceof ProtocolAdministrationData) {
			ProtocolAdministrationData pad = (ProtocolAdministrationData)pr;
			_scriptName = pad.scriptName;

			_frame = new JFrame(getTitle());
			JFrame.setDefaultLookAndFeelDecorated(true);
			_frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

			String columns[] = new String[pad.columns.length + 3];
			columns[0] = "#";
			columns[1] = "Zeitstempel";
			int i = 2;
			for(Column c : pad.columns) {
				columns[i++] = c.getCaption();
			}
			columns[columns.length - 1] = "Status";

			_model = new ProtocolTableModel(columns, _dataModel);

			JTable table = new JTable(_model);
			table.setPreferredScrollableViewportSize(new Dimension(1024, 400));
			table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);

			TableColumn column = null;
			for(int j = 0; j < table.getColumnCount(); j++) {
				column = table.getColumnModel().getColumn(j);
				if(j == 0) {
					column.setPreferredWidth(50);
				}
				else {
					column.setPreferredWidth(150);
				}
			}

			JScrollPane scrollPane = new JScrollPane(table);
			this.add(scrollPane);
			this.setOpaque(true); //content panes must be opaque
			_frame.setContentPane(this);

			_frame.pack();
			_frame.setVisible(true);
		}
		else if(_model != null) {
			if(pr instanceof ProtocolAggregationData) {
				ProtocolAggregationData pad = (ProtocolAggregationData)pr;
				_model.addAggregationInfo(pad);
			}
			else if(pr == null) {
				_frame.setTitle(_frame.getTitle() + "   _-_ PE abgeschlossen");
			}
			else {
				_model.add(pr);
			}
		}
	}

	/** @return Fenstertitel */
	private String getTitle() {
		return ((_scriptName != null) ? _scriptName : "") + ": " + ((_protocolId != null) ? _protocolId : "") + " - " + _date;
	}

	/**
	 * Setzt die Protokoll-Id. Wird in den Fensternamen aufgenommen.
	 *
	 * @param protocolId Protokoll-Id
	 */
	public void setProtocolId(long protocolId) {
		this._protocolId = protocolId;
		if(_frame != null) _frame.setTitle(getTitle());
	}

	/**
	 * Zeigt eine Fehlermeldung an
	 *
	 * @param errorMsg Fehlermeldung.
	 */
	public void showErrorMessage(String errorMsg) {
		if(_frame != null) {
			JOptionPane.showMessageDialog(_frame, errorMsg, "Es ist ein Fehler aufgetreten", JOptionPane.ERROR_MESSAGE);
		}
		else {
			System.out.println(errorMsg);
		}
	}

	/**
	 * Zeigt an ob das Fenster bereits erzeugt wurde.
	 *
	 * @return <code>true</code> falls das Fenster bereits aufgebaut wurde.<code>false</code> sonst.
	 */
	public boolean isReady() {
		return _frame != null;
	}

	/** @see java.awt.Window#dispose() */
	public void dispose() {
		if(_frame != null) _frame.dispose();
	}
}

/**
 * Verwaltet die Protokolldatensätze in einer Tabelle.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 */
class ProtocolTableModel extends AbstractTableModel {

	private static final long serialVersionUID = 5907963663506354756L;

	/** Kopfzeile */
	String header[];

	/** Werte */
	ArrayList<ProtocolResult> rows = new ArrayList<ProtocolResult>(500);

	/** Aggregationsdatensatz */
	private ProtocolAggregationData pad = null;

	private DataModel _dataModel;

	/** @param header Überschriften der Kopfzeile
	 * @param dataModel*/
	public ProtocolTableModel(String header[], final DataModel dataModel) {
		this.header = header;
		_dataModel = dataModel;
	}

	/**
	 * Fügt Aggregationsinformationen über das Protokoll hinzu. Diese werden in der letzten Tabellenzeile angezeigt.
	 *
	 * @param protocolAggregationData
	 */
	public void addAggregationInfo(ProtocolAggregationData protocolAggregationData) {
		this.pad = protocolAggregationData;
	}

	/** @see javax.swing.table.TableModel#getRowCount() */
	public int getRowCount() {
		return rows.size() + ((pad != null) ? 1 : 0);
	}

	/** @see javax.swing.table.TableModel#getColumnCount() */
	public int getColumnCount() {
		return header.length;
	}

	/** @see javax.swing.table.TableModel#getValueAt(int,int) */
	public Object getValueAt(int rowIndex, int columnIndex) {
		if(columnIndex == 0) {
			return rowIndex + 1;
		}
		else if(rowIndex < rows.size()) {
			if(rows.get(rowIndex) instanceof ProtocolResultData) {
				ProtocolResultData row = (ProtocolResultData)rows.get(rowIndex);

				if(columnIndex == 1) {
					return Util.msToDate(row.getTimeStamp());
				}
				else if(columnIndex == header.length - 1) {
					return row.status;
				}
				else {
					//nicht aggregierte Daten
					columnIndex = columnIndex - 2;
					if(columnIndex >= row.values.length) {
						return "-";
					}
					else {
						return row.getData(_dataModel, columnIndex).valueToString();
					}
				}
			}
			else if(rows.get(rowIndex) instanceof ProtocolAggregationResultData) {
								ProtocolAggregationResultData row = (ProtocolAggregationResultData)rows.get(rowIndex);

				if(columnIndex == 1) {
					return Util.msToDate(row.getStartTimeStamp()) + "-" + Util.msToDate(row.getEndTimeStamp());
				}
				else if(columnIndex == header.length - 1) {
					return row.order;
				}
				else {
					//nicht aggregierte Daten
					columnIndex = columnIndex - 2;
					if(columnIndex >= row.values.length) {
						return "-";
					}
					else {
						return row.getData(_dataModel, columnIndex).valueToString();
					}
				}
			}
			else if(rows.get(rowIndex) instanceof ProtocolNoChanges) {
				ProtocolNoChanges pnc = (ProtocolNoChanges)rows.get(rowIndex);
				if(columnIndex == 1) {
					if(pnc.order != ApplyAggregation.LISTE) {
						return pnc.order.toString();
					}
					else {
						return Util.msToDate(pnc.timestamp);
					}
				}
				else {
					return "\"";
				}
			}
			else {
				return "?";
			}
		}
		else	//letzte Zeile: Aggregationsdateninfo
		{
			if(columnIndex > 1) {
				columnIndex -= 2;
				final List<List<String>> aggregations = pad.getAggregations();
				if(columnIndex < aggregations.size()) {
					return aggregations.get(columnIndex) + ((pad.linkedAttributes != null) ? " (" + pad.linkedAttributes[columnIndex] + ")" : "");
				}
			}
			return "";
		}
	}

	/**
	 * Protokolldatensatz hinzufügen.
	 *
	 * @param pr Protokolldatensatz.
	 */
	public void add(ProtocolResult pr) {
		//Sicherstellen, dass nur ProtocolResultData und ProtocolNoChanges in rows landen
		if(pr instanceof ProtocolResultData || pr instanceof ProtocolAggregationResultData || pr instanceof ProtocolNoChanges) rows.add(pr);
//		this.fireTableRowsInserted(rows.size()-2, rows.size()-1);
		fireTableDataChanged();
	}

	/** @see javax.swing.table.TableModel#getColumnName(int) */
	@Override
    public String getColumnName(int col) {
		return header[col];
	}
}
