/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.interpreter.semantics;

import antlr.Token;
import antlr.collections.AST;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.util.ExpressionTree;
import de.bsvrz.pua.prot.util.MemberCheck;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.sys.funclib.losb.datk.NoChangeMarker;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.sys.funclib.losb.util.Tuple;

import java.util.*;

/**
 * Diverse Semantiküberprüfungen.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class MiscellaneousSemantics implements SemanticChecks {

	private final MemberCheck _check;

	private final SymbolTable _symbolTable;

	private final Semantics _semantics;

	/** Enthält die freien Aspekte, für die keine Bindung existiert */
	private final HashSet<String> noFreeAspects = new HashSet<String>();

	/** @param semantics Semantics  */
	public MiscellaneousSemantics(final Semantics semantics) {
		this._semantics = semantics;
		this._check = semantics.getCheck();
		this._symbolTable = semantics.symbolTable;
	}

	/** Fügt das Zeitdauer-Attribut in die Liste der im Spaltenbereich definierten Attribute ein. */
	public void addDurationAttribute() {
		_semantics.pi.addElement(_symbolTable._durationAttribute);
	}

	/**
	 * Fügt Objekte, für die das Protokoll erstellt werden soll, hinzu. Enthält die Liste nur ein Element, so handelt es sich dabei um ein Objekt, das hinzugefügt
	 * werden soll. Enthält pidList mehr als ein Element, so handelt es sich bei dem ersten Element um ein Objekt, und bei den übrigen Elementen um Mengen, wobei
	 * jede Menge immer die nachfolgende Menge enthält. Die Objekte der letzten Menge werden hinzugefügt.
	 *
	 * @param token   Token (Nur für die Zeilenangabe bei den Fehlermeldungen nötig)
	 * @param pidList Liste des Objektes und der Mengen.
	 *
	 * @throws CriticalParserException Falls es zu einem Fehler kommt
	 */
	public void addStandardObjects(final Token token, final List<String> pidList) throws CriticalParserException {
		try {
			if(pidList != null && !pidList.isEmpty()) {
				final String pid = pidList.remove(0);

				final SystemObject so = _check.getObject(pid);

				if(so == null) throw new SemanticErrorException(InterpreterErrorMessage.INVALID_OBJECT_PID + pid, token);

				if(!pidList.isEmpty()) {
					throw new SemanticErrorException(InterpreterErrorMessage.INVALID_OBJECT_SET, token);
//					if(so instanceof ConfigurationObject) {
//						try {
//							// aus dem ersten Objekt die referenzierte Menge holen
//							ConfigurationObject co = (ConfigurationObject)so;
//
//							pid = pidList.remove(0);
//              			ObjectSet set = null;
//							set = co.getObjectSet(pid);
//
//							if(set == null) throw new SemanticErrorException(InterpreterErrorMessage.IS_NO_SET + pid + ", " + so.getPid(), token);
//
//							// die Mengen durchlaufen
//							for(String element : pidList) {
//								pid = set.getPid();
//								set = set.getObjectSet(element);
//								if(set == null) throw new SemanticErrorException(InterpreterErrorMessage.IS_NO_SET + element + ", " + pid, token);
//							}
//
//							// alle Objekte aus dem set auslesen
//							List<SystemObject> elements = set.getElements();
//							if(elements != null) {
//								for(SystemObject sObj : elements) {
//									if(MemberCheck.whatIs(sObj) == MemberCheck.OBJECT) _symbolTable.addStandardObject(sObj);
//								}
//							}
//						}
//						catch(ConfigurationException ce) {
//							throw new SemanticErrorException(ce.getMessage(), token);
//						}
//					}
//					else {
//						throw new SemanticErrorException(InterpreterErrorMessage.OBJECT_W_SET_EXPECTED + pid, token);
//					}
				}
				else {
					// so MUSS ein Objekt sein - kein TYP!
					if(MemberCheck.whatIs(so) == MemberCheck.OBJECT) {
						_symbolTable.addStandardObject(so);
					}
					else {
						throw new SemanticErrorException(
								InterpreterErrorMessage.OBJECT_EXPECTED + so.getPid() + " (" + MemberCheck.toString(MemberCheck.whatIs(so)) + ")", token
						);
					}
				}
			}
		}
		catch(CriticalParserException criticalException) {
			criticalException.line = token.getLine();
			throw criticalException;
		}
	}

	/**
	 * Überprüft ob ein Bereich bereits definiert wurde oder ob bestimmte Schlüsselwörter mehrmals verwendet weden.
	 *
	 * @param token Token (Nur für die Zeilenangabe bei den fehlermeldungen nötig)
	 * @param field Bereich, der überprüft werden soll
	 * @param index Index, der überprüft wird
	 *
	 * @throws SemanticErrorException Falls Bereiche oder bestimmte Schlüsselwörter mehrfach verwendet werden.
	 */
	public void check(final Token token, final boolean[] field, final int index) throws SemanticErrorException {
		try {
			boolean result = false;

			if(index >= 0 && index <= field.length) {
				result = field[index];
			}

			if(result) {
				throw new SemanticErrorException(InterpreterErrorMessage.REDEFENITION);
			}
			else {
				field[index] = true;
			}
		}
		catch(SemanticErrorException exception) {
			exception.line = token.getLine();
			throw exception;
		}
	}

	/**
	 * Liefert den zu <code>placeholder</code> passenden, gebundenen Aspekt (als Aspekt-Pid) zurück
	 *
	 * @param placeholder Platzhalter des Aspekts
	 *
	 * @return Aspekt, der an den Platzhalter gebunden ist. Falls keine Bindung existiert, wird der Platzhalter zurückgegeben.
	 */
	public String getBinding(final String placeholder) {
		String result = _symbolTable.getBinding(placeholder);

		if(result == null) {
			noFreeAspects.add(placeholder);
			result = placeholder;
		}

		return result;
	}

	/**
	 * Setzt die 'Einschränkung'sanweisung
	 *
	 * @param expression Ausdruck
	 * @param merge      True falls verschmolzen werden soll, sonst false
	 */
	public void setConstraint(final ExpressionTree expression, final boolean merge) {
		_semantics.pi.setConstraint(new Tuple<ExpressionTree, Boolean>(expression, merge));
	}

	/**
	 * Legt den Nachfilter fest
	 *
	 * @param ast        Knoten (wird nur für Zeilenangabe in der Fehlermeldung benötigt)
	 * @param postFilter Klassenname des nachfilters
	 *
	 * @throws SemanticErrorException Falls es zu einem Fehler kommt, z.B. weil die Nachfilterklasse nicht existiert.
	 */
	public void setPostFilter(final AST ast, final String postFilter) throws SemanticErrorException {
		final String error = _semantics.pi.setPostFilter(postFilter);
		if(error != null) throw new SemanticErrorException(error, ast);
	}

	/**
	 * Setzt den Typ des Protokolls
	 *
	 * @param type True für Änderungs-, false für ein Statusprotokoll
	 */
	public void setProtocolType(final ProtocolType type) {
		_semantics.pi.setProtocolType(type);
	}


	/**
	 * Setzt die Übertragungsart von "Keine Änderungen"-Datensätzen
	 * @param noChangeMarker Übertragungsart von "Keine Änderungen"-Datensätzen
	 */
	public void setNoChangeMarker(final NoChangeMarker noChangeMarker) {
		_semantics.pi.setNoChangeMarker(noChangeMarker);
	}

	/**
	 * Abschließende Semantiküberprüfung. Stellt sicher, dass Bindungen zu allen freien Aspekten existieren. (Nur, falls processScript == true)
	 *
	 * @param pp            Wird nicht berücksichtigt.
	 * @param processScript True falls die Daten vollständig zur Datenaufbereitung sein müssen.
	 *
	 * @throws SemanticErrorException Es gibt noch ungebundene Aspekte.
	 */
	public void cleanUp(
			final ProcessingParameter pp, final boolean processScript) throws SemanticErrorException {
		// Bindungen zu allen freien Aspekten existieren
		if(!noFreeAspects.isEmpty() && processScript) {
			final StringBuffer errorMsg = new StringBuffer();
			errorMsg.append(InterpreterErrorMessage.NO_BINDING_FOR_ASPECT);
			for(final String fAsp : noFreeAspects) {
				errorMsg.append(", " + fAsp);
			}
			throw new SemanticErrorException(errorMsg.toString());
		}
	}

}
