/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.manager;

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.pua.prot.manager.datamanager.DataManager;
import de.bsvrz.pua.prot.manager.status.DavPublisher;
import de.bsvrz.pua.prot.manager.taskmanager.TaskManager;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.AtgProtocolRequest;
import de.bsvrz.sys.funclib.losb.datk.PidProtocol;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager;
import de.bsvrz.sys.funclib.losb.kernsoftware.SimpleSender;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

import java.util.Arrays;

/**
 * Klasse die Protokoll-bezogene Anfragen vom Datenverteiler entgegennimmt. Diese werden an den {@link de.bsvrz.pua.prot.manager.datamanager.DataManager} oder
 * {@link de.bsvrz.pua.prot.manager.taskmanager.TaskManager} weiterleitet.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class ProtocolRequestReceiver implements ClientReceiverInterface {

	private DataManager _dataManager;

	private TaskManager _taskManager;

	private ClientDavInterface _dav;

	private DataModel _model;

	private SystemObject _object;

	/** Synchronisationsobjekt. */
	private Object _subscription = new Object();

	/** Erster Datensatz, der nach Anmeldung als Senke eingeht. */
	private ResultData _firstData = null;

	/** Debug Ausgaben */
	private static final Debug _debug = Debug.getLogger();


	/**
	 * @param dav         Verbindung zum Datenverteiler.
	 * @param dataManager Data-Manager.
	 * @param taskManager Task-Manager.
	 * @param object      Objekt auf das sich der Receiver anmeldet.
	 *
	 * @throws FailureException       Fehler beim Anmelden als Empfänger.
	 * @throws ConfigurationException Fehler bei der Kommunikation mit der Konfiguration.
	 */
	public ProtocolRequestReceiver(ClientDavInterface dav, DataManager dataManager, TaskManager taskManager, SystemObject object)
			throws ConfigurationException, FailureException {
		this._dataManager = dataManager;
		this._taskManager = taskManager;
		this._dav = dav;
		this._model = dav.getDataModel();
		this._object = object;

		AttributeGroup protocolRequestAtg = _model.getAttributeGroup(PidProtocol.atgProtocolRequest);
		Aspect protocolRequestAsp = _model.getAspect(PidProtocol.aspRequest);
		DataDescription dd = new DataDescription(protocolRequestAtg, protocolRequestAsp);
		ConnectionManager.subscribeReceiver(dav, this, object, dd, ReceiveOptions.normal(), ReceiverRole.drain());
		waitForOk();
	}

	/**
	 * Wartet bis der Datenverteiler die Senken-Anmeldung bestätigt.
	 *
	 * @throws FailureException Senke-Anmeldung ist fehlgeschlagen
	 */
	private void waitForOk() throws FailureException {
		try {
			synchronized(_subscription) {
				while(_firstData == null) {
					_subscription.wait();
				}
			}
			if(_firstData.isNoValidSubscription()) throw new FailureException(ManagerErrorMessage.CA_IN_USE, LoggerException.ERROR);
		}
		catch(InterruptedException e) {
			;
		}
	}

	/** @see ClientReceiverInterface#update(ResultData[]) */
	public void update(ResultData[] results) {
		if(results != null) {
			synchronized(_subscription) {
				if(_firstData == null && results.length > 0) {
					_firstData = results[0];
					_subscription.notifyAll();
				}
			}
			for(ResultData result : results) {
				if(result.getData() != null) update(result);
			}
		}
	}

	/** @param result Auftragsparameter .Data von results darf nicht null sein! */
	private void update(ResultData result) {
		_debug.info("Protokollbezogene Anfrage angenommen.");
		Data data = result.getData();

		AtgProtocolRequest atgRequest = AtgProtocolRequest.getJavaObject(data);

		_debug.info("Operation der Anfrage: " + atgRequest.opCode);

		switch(atgRequest.opCode) {
			case ProtocolRelatedOperation.PROTOCOL_COMPLETION_INTERFACE:
				if(atgRequest.requestData.length >= 1 && atgRequest.client != null){
					if(atgRequest.requestData[0] == 3 && atgRequest.requestData.length >= 2){
						// Erweiterte Anmeldung mit Versionsnummer
						DavPublisher.addClient(_dav, atgRequest.client, atgRequest.requestData[1]);
						return;
					}		
					if(atgRequest.requestData[0] == 2){
						// Erweiterte Anmeldung
						DavPublisher.addClient(_dav, atgRequest.client, 2);
						return;
					}
					if(atgRequest.requestData[0] == 1){
						// Normale Anmeldung
						DavPublisher.addClient(_dav, atgRequest.client, 1);
						return;
					}
					if(atgRequest.requestData[0] == 0){
						// Abmeldung
						DavPublisher.removeClient(atgRequest.client);
						return;
					}
				}
				_debug.warning("Ungültige Operation: "  + atgRequest.opCode + " " + Arrays.toString(atgRequest.requestData));
				break;
			case ProtocolRelatedOperation.ABORT:
			case ProtocolRelatedOperation.ACK_FLAG: // Protokollübertragung anhalten ohne das Speichern des Protokolls abzubrechen
			case ProtocolRelatedOperation.CREATE:
			case ProtocolRelatedOperation.STATUS:
			case ProtocolRelatedOperation.PAUSE_PROTOCOL:
			case ProtocolRelatedOperation.RESUME_PROTOCOL:
			case ProtocolRelatedOperation.SEND_SAVED_PROTOCOL:
			case ProtocolRelatedOperation.SEND_UNREAD_PROTOCOL:
			case ProtocolRelatedOperation.PUA_ARS_STATUS:
			case ProtocolRelatedOperation.JOB_LIST:
			case ProtocolRelatedOperation.FLOW_CONTROL:
				long initialFlowControl = 0;
				try {
					initialFlowControl = data.getUnscaledValue("Flusskontrolle").longValue();
				}
				catch(Exception e){
					_debug.fine("Datenmodell unterstützt keine Flusskontrolle", e);
				}
				_taskManager.process(atgRequest.client, atgRequest.requestId, atgRequest.opCode, atgRequest.requestData, initialFlowControl);
				break;
			case ProtocolRelatedOperation.DELETE:
			case ProtocolRelatedOperation.LIST:
			case ProtocolRelatedOperation.UNREAD_PROTOCOL_LIST:
				//OPT momentan werden die infos client, requestId... mehrfach aus resultData geholt -> addProtocolRequest anpassen!
				//z.B: dataManager.addProtocolRequest(atgRequest.client, atgRequest.requestId, atgRequest.opCode, atgRequest.requestData);
				_dataManager.addProtocolRequest(new ResultData[]{result});
				break;
			default:
				if(atgRequest.client != null) {
					try {
						DataDescription dd = new DataDescription(
								_model.getAttributeGroup(PidProtocol.atgProtocolRequest), _model.getAspect(PidProtocol.aspAnswer)
						);
						SimpleSender.source(
								_dav, atgRequest.client, dd, AtgProtocolRequest.build(
								_dav, atgRequest.client, atgRequest.requestId, atgRequest.opCode, ManagerErrorMessage.INVALID_OPCODE + atgRequest.opCode
						)
						);
					}
					catch(LoggerException e) {
						e.log();
					}
				}
		}
	}

	/** Meldet den Empfang von Protokoll-bezogenen Anfragen ab. */
	public void unsubscribe() {
		try {
			AttributeGroup protocolRequestAtg = _model.getAttributeGroup(PidProtocol.atgProtocolRequest);
			Aspect protocolRequestAsp = _model.getAspect(PidProtocol.aspRequest);
			DataDescription dd = new DataDescription(protocolRequestAtg, protocolRequestAsp);
			ConnectionManager.unsubscribeReceiver(_dav, this, _object, dd);
		}
		catch(Exception e) {
			_debug.error(ErrorMessage.COMMUNICATION, e);
		}
	}
}
