/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.manager.datamanager;

import de.bsvrz.pua.prot.client.dataobject.ProtocolResult;
import de.bsvrz.pua.prot.exceptions.NoncriticalException;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;

/**
 * Klasse zum Zugriff auf gespeicherte Protokolle.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ProtocolInputStream {

	/** Debug Ausgabe */
	private static final Debug _debug = Debug.getLogger();

	/** Stream auf Protokolldatei. */
	private ObjectInputStream _ois = null;

	/** Verzeichnis, in dem die Protokolldatei abgelegt ist. */
	private final String _path;

	/** Protokoll-Id */
	private final long _protocolId;

	/** Wurde das Protokoll bereits abgerufen? */
	private final boolean _alreadyRead;

	/** Dateiname des Protokolls. */
	File _file;

	/**
	 * Öffnet eine Protokolldatei.
	 *
	 * @param protocolId protocolId für die Datei, in die geschrieben werden soll.
	 * @param path       Verzeichnis, in dem das Protokoll abgelegt werden soll. Muss auf {@link File#separator} enden.
	 * @throws FailureException Fehler beim Zugriff auf die Datei.
	 */
	public ProtocolInputStream(long protocolId, String path) throws FailureException {
		this._path = path;
		this._protocolId = protocolId;
		File readFile = new File(path, protocolId + (ProtocolOutputStream.PROTOCOL_FILENAME_EXTENSION));
		File unreadFile = new File(path , protocolId + (ProtocolOutputStream.PROTOCOL_NOT_READ_YET));
		if(unreadFile.exists()) {
			_file = unreadFile;
			_alreadyRead = false;
		}
		else {
			_file = readFile;
			_alreadyRead = true;
		}

		FileInputStream fis = null;
		try {
			fis = new FileInputStream(_file);
		}
		catch(FileNotFoundException e) {
			throw new FailureException("Protokoll nicht vorhanden: " + _protocolId, e, LoggerException.WARNING);
		}
		try {
			_ois = new ObjectInputStream(fis);

			//Auftragsparameter werden ignoriert.
			Object in = _ois.readObject();
			if(in == null) {
				throw new FailureException(ErrorMessage.NO_DATA + _file, LoggerException.WARNING);
			}
			else if(!(in instanceof byte[]))    //Die ProcessingParameter werden als ByteStream abgelegt!
			{
				throw new FailureException(
						ErrorMessage.UNEXPECTED_DATA_TYPE + in.getClass().getName() + ". " + _file + " ist keine Protokolldatei.", LoggerException.WARNING
				);
			}
		}
		catch(IOException e) {
			throw new FailureException(ErrorMessage.CAN_NOT_ACCESS_FILE + _file + ": " + e.getMessage(), e, LoggerException.WARNING);
		}
		catch(ClassNotFoundException e) {
			throw new FailureException(e, LoggerException.WARNING);
		}
	}

	/**
	 * Liefert den nächsten Datensatz, der im Protokoll gespeichert ist.
	 *
	 * @return Datensatz oder <code>null</code> wenn alle Datensätze abgefragt wurden.
	 *
	 * @throws FailureException Fehler beim Zugriff auf die protokolldatei.
	 */
	public ProtocolResult take() throws FailureException {
		try {
			if(_ois != null) {
				try {
					Object in = _ois.readObject();
					if(in instanceof ProtocolResult) {
						return (ProtocolResult)in;
					}
					else {
						throw new FailureException(
								ErrorMessage.UNEXPECTED_DATA_TYPE + in.getClass().getName() + ". " + _file + " ist keine Protokolldatei.",
								LoggerException.WARNING
						);
					}
				}
				catch(EOFException e) {
					close();
					return null;
				}
			}
			else {
				return null;
			}
		}
		catch(ClassNotFoundException e) {
			throw new FailureException(e, LoggerException.WARNING);
		}
		catch(IOException e) {
			throw new FailureException(ErrorMessage.CAN_NOT_ACCESS_FILE + _file + ": " + e.getMessage(), e, LoggerException.WARNING);
		}
	}

	/** Beendet den Zugriff auf das Protokoll. Muss aufgerufen werden, wenn keine weiteren Zugriffe mehr auf die Protokolldatei erfolgen. */
	public void close() {
		try {
			if(_ois != null) {
				_ois.close();
				_ois = null;
			}
		}
		catch(IOException e) {
			_debug.warning(ErrorMessage.CAN_NOT_ACCESS_FILE + e.getMessage());
		}
	}

	/**
	 * Beendet den Zugriff auf das Protokoll und nennt das Protokoll in ein gespeichertes Protokoll um. Muss aufgerufen werden, wenn keine weiteren Zugriffe mehr
	 * auf die Protokolldatei erfolgen.
	 * @throws NoncriticalException Exception
	 */
	public void closeAndRename() throws NoncriticalException {
		close();
		if(!_alreadyRead) {
			ProtocolOutputStream.renameProtocol(_file, new File(_path, _protocolId + ProtocolOutputStream.PROTOCOL_FILENAME_EXTENSION));
		}
	}
}
