/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung - Bibliothek
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.lib;

import java.text.DateFormat;
import java.util.Date;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.puk.param.lib.daten.DataWithTime;

/**
 * die Repräsentation eines Parameters für eine Parameteridentifikation.<br>
 * Die Klasse vereint die Informationen über die Spezifikation des Parameters,
 * seine Daten und die Quelle, aus der die Daten ermittelt wurden.
 *
 * @author BitCtrl Systems GmbH, Peuker
 * @version $Id$
 */
public class Parameter {

	/** die Parameterspezifikation. */
	private final ParameterInfo info;

	/** die Daten des Parameters. */
	private DataWithTime data;

	/** die Quelle, von der die Daten ermittelt wurden. */
	private Quelle quelle;

	/**
	 * erzeugt einen leeren Parameterdatensatz für die übergebene
	 * Parameterspezifikation.
	 *
	 * @param info
	 *            die Parameterspezifikation
	 */
	public Parameter(final ParameterInfo info) {
		this.info = info;
		quelle = new Quelle(info.getObjekt(), QuellTyp.LOKAL_OBJEKT);
	}

	/**
	 * erzeugt einen Parameterdatensatz mit den übergebenen Datenspezifikation
	 * und den übergebenen Daten.
	 *
	 * @param parameter
	 *            die Spezifikation
	 * @param data
	 *            die Daten
	 *
	 * @deprecated die Parameter sollen mit Zeitstempel gespeichert und
	 *             veröffentlichet werden, daher sollte der Konstruktor
	 *             {@link #Parameter(ParameterInfo, DataWithTime)} verwendet
	 *             werden, wenn Daten zugeweisen werden.
	 */
	@Deprecated
	public Parameter(final ParameterInfo parameter, final Data data) {
		this(parameter);
		setDataWithTime(new DataWithTime(data, 0));
		if (parameter.getObjekt().getType().isOfType(parameter.getTyp())) {
			quelle = new Quelle(info.getObjekt(), QuellTyp.LOKAL_OBJEKT);
		} else if (parameter.getObjekt() instanceof SystemObjectType) {
			quelle = new Quelle(info.getObjekt(), QuellTyp.LOKAL_TYP);
		} else {
			quelle = new Quelle(info.getObjekt(), QuellTyp.LOKAL_HIERARCHIE);
		}
	}

	/**
	 * erzeugt einen Parameterdatensatz mit den übergebenen Datenspezifikation
	 * und den übergebenen Daten und dem zugeprdneten Zeitpunkt.
	 *
	 * Haben die übergebenen Daten einen Zeitpunkt mit einem Wert &lt;= 0, wird
	 * die aktuelle Systemzeit verwendet.
	 *
	 * @param parameter
	 *            die Spezifikation
	 * @param data
	 *            die Daten
	 */
	public Parameter(final ParameterInfo parameter, final DataWithTime data) {
		this(parameter);
		setDataWithTime(data);
		if (parameter.getObjekt().getType().isOfType(parameter.getTyp())) {
			quelle = new Quelle(info.getObjekt(), QuellTyp.LOKAL_OBJEKT);
		} else if (parameter.getObjekt() instanceof SystemObjectType) {
			quelle = new Quelle(info.getObjekt(), QuellTyp.LOKAL_TYP);
		} else {
			quelle = new Quelle(info.getObjekt(), QuellTyp.LOKAL_HIERARCHIE);
		}
	}

	/**
	 * liefert die Attrbutgruppe die die Daten des Parameters beschreibt.
	 *
	 * @return die Attributgruppe
	 */
	public AttributeGroup getAtg() {
		return info.getAtg();
	}

	/**
	 * liefert die Daten des Parameters.
	 *
	 * @return die Daten oder <i>null</i>, wenn keine verfügbar sind
	 */
	public Data getData() {
		if (data == null) {
			return null;
		}
		return data.getData();
	}

	/**
	 * liefert den Zeitstempel der Daten.
	 *
	 * @return die Daten oder <i>0</i>, wenn keine verfügbar sind
	 */
	public long getZeitpunkt() {
		if (data == null) {
			return 0;
		}
		return data.getZeitpunkt();
	}

	/**
	 * liefert die Daten des Parameters mit zugeordnetem Zeitstempel.
	 *
	 * @return die Daten oder <i>null</i>, wenn keine verfügbar sind
	 */
	public DataWithTime getDataWithTime() {
		return data;
	}

	/**
	 * liefert die Sepzifikation des Parameters.
	 *
	 * @return die Spezifikation
	 */
	public ParameterInfo getInfo() {
		return info;
	}

	/**
	 * liefert das Objekt, dessen Parameter beschrieben werden.
	 *
	 * @return das Objekt
	 */
	public SystemObject getObjekt() {
		return info.getObjekt();
	}

	/**
	 * liefert die Quelle, aus denen die Daten des Parameters ermittelt wurden.
	 *
	 * @return die Quelle
	 */
	public Quelle getQuelle() {
		return quelle;
	}

	/**
	 * liefert eine serialisierbare Repräsentation des Parameters, die in einem
	 * Bytearray in einer Datenverteiler-Attributgruppe versendet werden kann.
	 *
	 * @param verbindung
	 *            die verwendete Datenverteilerverbindung
	 * @return das serialisierbare Objekt
	 */
	public SerializableParameter getSerializableParameter(final ClientDavInterface verbindung) {
		return new SerializableParameter(verbindung, this);
	}

	/**
	 * liefert die Simulationsvariante, für die die Daten ermittelt wurden.
	 *
	 * @return die Simulationsvariante
	 */
	public short getSim() {
		return info.getSim();
	}

	/**
	 * liefert den Typ des Objekts für den der Parameter bestimmt wurde.
	 *
	 * @return den Typ
	 */
	public SystemObjectType getTyp() {
		return info.getTyp();
	}

	/**
	 * ermittelt, ob Daten für den Parameter definiert sind.
	 *
	 * @return <i>true</i>, wenn der Parameter Daten enthält.
	 */
	public boolean hasData() {
		return null != data;
	}

	/**
	 * die Funktion ermittelt, ob die Quelle Daten liefern kann.
	 *
	 * @return es können potentiell Daten bestimmt werden
	 */
	public boolean isDataAvailable() {
		return hasData() || quelle.isDataAvailable();
	}

	/**
	 * setzt die Daten des Parameters.
	 *
	 * @param data
	 *            die Daten
	 *
	 * @deprecated die Parameter sollten einen zugeordneten Zeitpunkt haben,
	 *             deshalb sollte die Funktion
	 *             {@link #setDataWithTime(DataWithTime)} verwendet werden.
	 */
	@Deprecated
	public void setData(final Data data) {
		setDataWithTime(new DataWithTime(data, 0));
	}

	/**
	 * setzt die Daten. Haben die Daten einen Zeitpunkt mit dem Wert &lt;= 0,
	 * wird die aktuelle Systemzeit verwendet.
	 *
	 * @param newData
	 *            die Daten
	 */
	public void setDataWithTime(final DataWithTime newData) {

		DataWithTime setData = newData;
		if (newData != null) {
			final long zeitpunkt = newData.getZeitpunkt();
			if (zeitpunkt <= 0) {
				setData = new DataWithTime(newData.getData(), System.currentTimeMillis());
			}

		}
		this.data = setData;
	}

	/**
	 * setzt die Quelle des Parameters.
	 *
	 * @param obj
	 *            das Objekt an dem die Parameter ermittelt wurden
	 * @param typ
	 *            der Typ der Quelle
	 */
	public void setQuelle(final SystemObject obj, final QuellTyp typ) {
		if (null == quelle) {
			quelle = new Quelle(obj, typ);
		} else {
			quelle.setObjekt(obj);
			quelle.setTyp(typ);
		}
	}

	@Override
	public String toString() {
		final StringBuffer buffer = new StringBuffer(100);
		buffer.append(info);
		buffer.append(", Quelle: ");
		buffer.append(quelle);
		if (data != null) {
			buffer.append(",  ");
			buffer.append(DateFormat.getDateTimeInstance().format(new Date(data.getZeitpunkt())));
			buffer.append(": Daten: ");
			buffer.append(data.getData());
		}

		return buffer.toString();
	}
}
